/*  gngb, a game boy color emulator
 *  Copyright (C) 2001 Peponas Thomas & Peponas Mathieu
 * 
 *  This program is free software; you can redistribute it and/or modify  
 *  it under the terms of the GNU General Public License as published by   
 *  the Free Software Foundation; either version 2 of the License, or    
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <SDL.h>
#include "global.h"
#include "memory.h"
#include "vram.h"
#include "message.h"
#include "tiny_font.c"
#include "emu.h"
#include "interrupt.h"
#include "sgb.h"
#include "video_std.h"

#ifdef DEBUG
#include "gngb_debuger/debuger.h"
#endif

static Uint32 std_flag;
static SDL_Surface *back=NULL;

Sint8 rb_tab[2][RB_SIZE]={{0,-2,2,-2,2},
			 {0,-1,-1,1,1}};

VIDEO_MODE video_std;

static __inline__ void draw_char_std(SDL_Surface *dest,int x,int y,unsigned char c) {
  static SDL_Rect font_rect,dest_rect;
  int indice=c-32;
  
  if (c<32 || c>127) return;
     
  font_rect.x=indice*wl;
  font_rect.y=0;
  font_rect.w=wl;
  font_rect.h=hl;
 
  dest_rect.x=x;
  dest_rect.y=y;
  dest_rect.w=wl;
  dest_rect.h=hl;

  SDL_BlitSurface(fontbuf,&font_rect,dest,&dest_rect);
}

void draw_message_std(int x,int y,char *mes) {
  int i;

  x+=scxoff;
  y+=scyoff;
    
  for(i=0;i<strlen(mes);i++)
    draw_char_std(gb_screen,x+i*wl,y,mes[i]);
}

void blit_screen_default_std(void) {
 
  if (rb_on) {
      rb_shift++;
      if (rb_shift>=RB_SIZE) {
	rb_on=0;
	rb_shift=0;
      }
    }

  scrR.x=0; scrR.y=0; scrR.w=SCREEN_X; scrR.h=SCREEN_Y;
  
  dstR.x=rb_tab[0][rb_shift]+scxoff; 
  dstR.y=rb_tab[1][rb_shift]+scyoff; 
  dstR.w=SCREEN_X; 
  dstR.h=SCREEN_Y;
  
  SDL_BlitSurface(back,&scrR,gb_screen,&dstR);
  
  update_message();
  SDL_Flip(gb_screen);
    
  if ((!(LCDCCONT&0x20) || !(LCDCCONT&0x01)) && (conf.gb_type&NORMAL_GAMEBOY)) clear_screen();
}




void blit_screen_sgb_std(void) {
  
  if (rb_on) {
      rb_shift++;
      if (rb_shift>=RB_SIZE) {
	rb_on=0;
	rb_shift=0;
      }
    }

  scrR.x=0; scrR.y=0; scrR.w=SCREEN_X; scrR.h=SCREEN_Y;
  
  dstR.x=rb_tab[0][rb_shift]+scxoff; 
  dstR.y=rb_tab[1][rb_shift]+scyoff; 
  dstR.w=SCREEN_X; 
  dstR.h=SCREEN_Y;
  
  /* FIXME: sgb support and inverse the blit*/
  /*    if (conf.gb_type&SUPER_GAMEBOY) 
	SDL_BlitSurface(sgb_buf,NULL,gb_screen,NULL);*/
  if (sgb_mask) SDL_FillRect(back,NULL,pal_bck[0]);
  else SDL_BlitSurface(back,&scrR,gb_screen,&dstR);
  
  update_message();
  SDL_Flip(gb_screen);
    
  if ((!(LCDCCONT&0x20) || !(LCDCCONT&0x01)) && (conf.gb_type&NORMAL_GAMEBOY))
    clear_screen();
}

void init_video_std(Uint32 flag) {
  std_flag=flag|SDL_HWSURFACE;

  if (conf.gb_type&SUPER_GAMEBOY)
    gb_screen=SDL_SetVideoMode(SGB_WIDTH,SGB_HEIGHT,
			       BIT_PER_PIXEL,std_flag);
  else 
    gb_screen=SDL_SetVideoMode(SCREEN_X,SCREEN_Y,
			       BIT_PER_PIXEL,std_flag); 
    back=SDL_CreateRGBSurface(SDL_HWSURFACE,SCREEN_X,SCREEN_Y+1,BIT_PER_PIXEL,
			    0xf800,0x7e0,0x1f,0x00);

    if (gb_screen==NULL) {
      printf("Couldn't set %dx%dx%d video mode: %s\n",
	     SCREEN_X,SCREEN_Y,BIT_PER_PIXEL,SDL_GetError());
      exit(1);
    }
    if (back==NULL) {
      printf("Couldn't allocate %dx%dx%d SDL_Surface: %s\n",
	     SCREEN_X,SCREEN_Y,BIT_PER_PIXEL,SDL_GetError());
      exit(1);
    }
    
    if (conf.gb_type&SUPER_GAMEBOY) {
      scxoff=48;
      scyoff=40;
      video_std.blit=blit_screen_sgb_std;
    } else {
      scxoff=0;
      scyoff=0;
      video_std.blit=blit_screen_default_std;
    }

}


/*void (*blit_screen_std)(void);*/

__inline__ void draw_spr_std(Uint16 *buf,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[0][sp->no_tile<<4];
  if (!sp->yflip) tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    if (c) {
      if (!(sp->priority)) buf[sp->x+nx]=grey[pal_obj[sp->pal][c]];
      else if (!back_col[sp->x+nx][CURLINE]) buf[sp->x+nx]=grey[pal_obj[sp->pal][c]];
    }
  }
}

__inline__ void draw_obj_std(Uint16 *buf)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) 
    draw_spr_std(buf,&gb_spr[(Uint8)i]);
}

__inline__ void draw_back_std(Uint16 *buf)
{
  Uint8 *tb,*tp;
  int y,x,i;
  int sx,sy;
  Uint8 /*bit0,bit1,*/c;
  Sint16 no_tile;
 
  if (LCDCCONT&0x08) tb=&vram_page[0][0x1c00];
  else  tb=&vram_page[0][0x1800];

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;

  no_tile=tb[i&0x1f];
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[0][no_tile<<4];
  tp+=(sy&0x07)<<1;
  
  x=0;
  for(sx=SCRX&0x07;sx<8;sx++,x++) {
    /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
      bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(sx);
    /* FIXME VIC VIPER */
    buf[x]=grey[gblcdc->vram_pal_line[x][c]];
    back_col[x][CURLINE]=c;
  }
  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    tp=&vram_page[0][no_tile<<4];
    tp+=(sy&0x07)<<1;
    for(sx=0;sx<8 && x+sx<160;sx++) {
      /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(sx);
      /* FIXME VIC VIPER */
      buf[x+sx]=grey[gblcdc->vram_pal_line[x+sx][c]];
      back_col[x+sx][CURLINE]=c;
    }
  }
}

__inline__ void draw_win_std(Uint16 *buf)
{
  Uint8 *tb,*tp;
  Sint16 y,x,i,sx=0;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c;
 
  if (LCDCCONT&0x40) tb=&vram_page[0][0x1c00];
  else tb=&vram_page[0][0x1800];
  
  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07)<<1;
    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[0][no_tile<<4];
      tp+=y;
      for(sx=0;sx<8 && (x+sx)<160;sx++) {
	/*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(sx);
	/* FIXME VIC VIPER */
	buf[x+sx]=grey[gblcdc->vram_pal_line[x+sx][c]];
	back_col[x+sx][CURLINE]=c;
      }
    }
  }  
}

/* CGB drawing functions */

__inline__ void draw_spr_col_std(Uint16 *buf,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[sp->page][sp->no_tile<<4];
  if (!sp->yflip) tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);

    /* TODO: OPTIMISATION */
    /*if (back_col[sp->x+nx][CURLINE]&0x80) {  
      if (c) {
	if (!(back_col[sp->x+nx][CURLINE]&0x0f)) 
	  buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      }
    } else {
      if (!sp->priority) {
	if (c) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      } else {
	if (!(back_col[sp->x+nx][CURLINE]&0x0f)) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      }      
      }*/

    if (LCDCCONT&0x01) {
      if (c) {
	if (back_col[sp->x+nx][CURLINE]&0x80) {
	  if (!(back_col[sp->x+nx][CURLINE]&0x0f)) 
	    buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
 	} else if (!sp->priority) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
	else if (!(back_col[sp->x+nx][CURLINE]&0x0f)) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      } 
    } else if (c) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
  }
}


__inline__ void draw_obj_col_std(Uint16 *buf)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) 
    draw_spr_col_std(buf,&gb_spr[(Uint8)i]);
}

__inline__ void draw_back_col_std(Uint16 *buf)
{
  Uint8 *tb,*tp,*att_tb;
  Uint16 y,x,i;
  Uint16 sx,sy;
  Uint8 /*bit0,bit1,*/c,p,att,xflip,yflip;
  Sint16 no_tile;
 
  if (LCDCCONT&0x08) {// select Tile Map
    tb=&vram_page[0][0x1c00];
    att_tb=&vram_page[1][0x1c00];
  } else {
    tb=&vram_page[0][0x1800];
    att_tb=&vram_page[1][0x1800];
  }

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  att_tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;
  
  no_tile=tb[i&0x1f];
  att=att_tb[i&0x1f];
  
  yflip = (att & 0x40) ? 1 : 0;
  xflip = (att & 0x20) ? 1 : 0;
  p=att&0x07;
   
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[(att&0x08)>>3][(no_tile<<4)];
  
  if (!yflip) tp+=((sy&0x07)<<1);
  else tp+=((7-(sy&0x07))<<1);

  x=0;
  for(sx=SCRX&0x07;sx<8;sx++,x++) {
    int wbit;
    if (!xflip) wbit=sx;
    else wbit=7-sx;
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    buf[x]=pal_col_bck[p][c];
    back_col[x][CURLINE]=(att&0x80)+c;
  }

  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    att=att_tb[i&0x1f];
    yflip = (att & 0x40) ? 1 : 0;
    xflip = (att & 0x20) ? 1 : 0;    
    p=att&0x07;

    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    
    tp=&vram_page[(att&0x08)>>3][(no_tile<<4)];
    if (!yflip) tp+=((sy&0x07)<<1);
    else tp+=((7-(sy&0x07))<<1);

    for(sx=0;sx<8;sx++) {
      int wbit;
      if (!xflip) wbit=sx;
      else wbit=7-sx;
      /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(wbit);
      buf[x+sx]=pal_col_bck[p][c];
      back_col[x+sx][CURLINE]=(att&0x80)+c;
    }
  }
}

__inline__ void draw_win_col_std(Uint16 *buf)
{
  Uint8 *tb,*tp,*att_tb; // ,*tiles;
  Uint16 y,x,i,sx;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c,p,att,xflip,yflip;
  
  if (LCDCCONT&0x40) {
    tb=&vram_page[0][0x1c00];
    att_tb=&vram_page[1][0x1c00];
  } else {
    tb=&vram_page[0][0x1800];
    att_tb=&vram_page[1][0x1800];
  }

  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    att_tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07);

    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      att=att_tb[i];
      yflip = (att & 0x40) ? 1 : 0;
      xflip = (att & 0x20) ? 1 : 0;  
      p=att&0x07;

      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[(att&0x08)>>3][(no_tile<<4)]; 
      if (!yflip) tp+=y<<1;
      else tp+=((7-y)<<1);
      
      for(sx=0;sx<8;sx++) {
	int wbit;
	if (!xflip) wbit=sx;
	else wbit=7-sx;
	/*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	  bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(wbit);
	buf[x+sx]=pal_col_bck[p][c];
	if (!(att&0x80)) back_col[x+sx][CURLINE]=c;
	else back_col[x+sx][CURLINE]=0x80+c;	    
      }
    }
  }  
}


/* SGB drawing Functions  To Optimize */

__inline__ void draw_back_sgb_std(Uint16 *buf)
{
  Uint8 *tb,*tp;
  int y,x,i;
  int sx,sy;
  Uint8 /*bit0,bit1,*/c;
  Sint16 no_tile;

  if (LCDCCONT&0x08) tb=&vram_page[0][0x1c00];
  else  tb=&vram_page[0][0x1800];

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;

  no_tile=tb[i&0x1f];
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[0][no_tile<<4];
  tp+=(sy&0x07)<<1;
  
  x=0;
  for(sx=SCRX&0x07;sx<8;sx++,x++) {
    /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
      bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
      c=(bit1<<1)|bit0;*/
     c=GET_GB_PIXEL(sx);
    //buf[x]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][pal_bck[c]];
    /* FIXME VIC VIPER */
    buf[x]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][gblcdc->vram_pal_line[x+sx][c]];
    back_col[x][CURLINE]=c;
  }
  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    tp=&vram_page[0][no_tile<<4];
    tp+=(sy&0x07)<<1;
    for(sx=0;sx<8 && x+sx<160;sx++) {
      /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(sx);
      //buf[x+sx]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][pal_bck[c]];
      //buf[x+sx]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][c];
      /* FIXME VIC VIPER */
      buf[x+sx]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][gblcdc->vram_pal_line[x+sx][c]];
      back_col[x+sx][CURLINE]=c;
    }
  }

}

__inline__ void draw_win_sgb_std(Uint16 *buf)
{
  Uint8 *tb,*tp;
  Sint16 y,x,i,sx;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c;
 
  if (LCDCCONT&0x40) tb=&vram_page[0][0x1c00];
  else tb=&vram_page[0][0x1800];
  
  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07)<<1;
    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[0][no_tile<<4];
      tp+=y;
      for(sx=0;sx<8 && x+sx<160;sx++) {
	/*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(sx);
	//buf[x+sx]=pal_bck[c];
	//buf[x+sx]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][pal_bck[c]];
	buf[x+sx]=sgb_pal[sgb_pal_map[x/8][CURLINE/8]][gblcdc->vram_pal_line[x+sx][c]];
	back_col[x+sx][CURLINE]=c;
      }
    }
  }  
}

__inline__ void draw_spr_sgb_std(Uint16 *buf,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[0][sp->no_tile<<4];
  if (!sp->yflip) 
    tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    if (c) {
      /* FIXME */
      //   printf("%d %d %d %d %d %d \n",sp->x,sp->y,CURLINE,sgb_pal_map[(sp->x+nx)/8][CURLINE/8],sp->pal,c);
      if (!(sp->priority)) buf[sp->x+nx]=sgb_pal[sgb_pal_map[(sp->x+nx)/8][CURLINE/8]][pal_obj[sp->pal][c]];
      else if (!back_col[sp->x+nx][CURLINE]) buf[sp->x+nx]=sgb_pal[sgb_pal_map[(sp->x+nx)/8][CURLINE/8]][pal_obj[sp->pal][c]];       
      
      /*if (!(sp->priority)) buf[sp->x+nx]=sgb_pal[0][pal_obj[sp->pal][c]];
	else if (!back_col[sp->x+nx][CURLINE]) buf[sp->x+nx]=sgb_pal[0][pal_obj[sp->pal][c]];*/
      /*if (!(sp->priority)) buf[sp->x+nx]=sgb_pal[sp->pal_col][c];
	else if (!back_col[sp->x+nx][CURLINE]) buf[sp->x+nx]=sgb_pal[sp->pal_col][c];*/
    }
  }
}

__inline__ void draw_obj_sgb_std(Uint16 *buf)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) 
    draw_spr_sgb_std(buf,&gb_spr[(Uint8)i]);
}

void draw_screen_wb_std(void)
{
  Uint16 *buf=(Uint16 *)back->pixels + CURLINE*SCREEN_X;
  if (SDL_MUSTLOCK(back) && SDL_LockSurface(back)<0) printf("can't lock surface\n");
#ifndef DEBUG
  if (LCDCCONT&0x01) draw_back_std(buf);
  if (LCDCCONT&0x20) draw_win_std(buf);
  if (LCDCCONT&0x02) draw_obj_std(buf);
#else
  if (active_back && LCDCCONT&0x01) draw_back_std(buf);
  if (active_win && LCDCCONT&0x20) draw_win_std(buf);
  if (active_obj && LCDCCONT&0x02) draw_obj_std(buf);
#endif
  if (SDL_MUSTLOCK(back)) SDL_UnlockSurface(back);
}

void draw_screen_col_std(void) 
{
  Uint16 *buf=(Uint16 *)back->pixels + CURLINE*SCREEN_X;
  if (SDL_MUSTLOCK(back) && SDL_LockSurface(back)<0) printf("can't lock surface\n");
#ifndef DEBUG
  draw_back_col_std(buf);
  if (LCDCCONT&0x20) draw_win_col_std(buf);
  if (LCDCCONT&0x02) draw_obj_col_std(buf);
#else
  if (active_back) draw_back_col_std(buf);
  if (active_win &&  LCDCCONT&0x20) draw_win_col_std(buf);
  if (active_obj && LCDCCONT&0x02) draw_obj_col_std(buf);
#endif
  if (SDL_MUSTLOCK(back)) SDL_UnlockSurface(back);
}

void draw_screen_sgb_std(void)
{
  Uint16 *buf=(Uint16 *)back->pixels + CURLINE*SCREEN_X;
  if (SDL_MUSTLOCK(back) && SDL_LockSurface(back)<0) printf("can't lock surface\n");
#ifndef DEBUG
  if (LCDCCONT&0x01) draw_back_sgb_std(buf);
  if (LCDCCONT&0x20) draw_win_sgb_std(buf);
  /* FIXME */
  if (LCDCCONT&0x02) draw_obj_sgb_std(buf);
#else
  if (active_back && LCDCCONT&0x01) draw_back_sgb_std(buf);
  if (active_win &&  LCDCCONT&0x20) draw_win_sgb_std(buf);
  /* FIXME */
  if (active_obj &&  LCDCCONT&0x02) draw_obj_sgb_std(buf);
#endif
  if (SDL_MUSTLOCK(back)) SDL_UnlockSurface(back);
}

void clear_screen_std(void)
{
  if (back)  SDL_FillRect(back,NULL,grey[0]);
}

VIDEO_MODE video_std={
  init_video_std,
  NULL,
  draw_screen_col_std,
  draw_screen_wb_std,
  draw_screen_sgb_std,
  NULL,
  NULL,
  draw_message_std,
  clear_screen_std
};



