/*  gngb, a game boy color emulator
 *  Copyright (C) 2001 Peponas Thomas & Peponas Mathieu
 * 
 *  This program is free software; you can redistribute it and/or modify  
 *  it under the terms of the GNU General Public License as published by   
 *  the Free Software Foundation; either version 2 of the License, or    
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <config.h>
#include <SDL.h>
#include "global.h"
#include "memory.h"
#include "video_yuv.h"
#include "vram.h"
#include "message.h"
#include "tiny_font.c"
#include "emu.h"
#include "interrupt.h"

#ifdef DEBUG
#include "gngb_debuger/debuger.h"
#endif

static __inline__ void draw_char_yuy2(int ox,int oy,unsigned char c) {
  int fx,fy;
  int indice=c-32;
  int x,y;
  Uint32 *buf=(Uint32 *)overlay->pixels[0] + oy*SCREEN_X;
  
  fx=indice*wl;
  fy=0;
  
  for(y=0;y<hl;y++,buf+=SCREEN_X-1) {
    for(x=0;x<wl;x++) {
      static Uint32 Y;
      static Uint32 U=128;
      Y=tiny_font.pixel_data[((x+fx)+(y+fy)*tiny_font.width)*3];
#ifndef GNGB_BIG_ENDIAN
      buf[y+ox+x]=(U<<24)|(Y<<16)|(U<<8)|Y;
#else
      buf[y+ox+x]=(Y<<24)|(U<<16)|(Y<<8)|U;
#endif
    }
  }
}

void draw_message_yuy2(int x,int y,char *mes) {
  int i;
  for(i=0;i<strlen(mes);i++)
    draw_char_yuy2(x+i*wl,y,mes[i]);
}



void init_video_yuy2(Uint32 flag)
{
  yuv_flag=SDL_HWSURFACE|flag|SDL_RESIZABLE; /* YUV mode are resizable */
  gb_screen=SDL_SetVideoMode(conf.res_w,conf.res_h,BIT_PER_PIXEL, yuv_flag);
  overlay=SDL_CreateYUVOverlay(SCREEN_X*2,SCREEN_Y,SDL_YUY2_OVERLAY,gb_screen); // FIXME : SGB support

  /* Error check */
  if (gb_screen==NULL) {
    printf("Couldn't set %dx%dx%d video mode: %s\n",
	   SCREEN_X,SCREEN_Y,BIT_PER_PIXEL,SDL_GetError());
    exit(1);
  }
  if (overlay==NULL) {
    printf("Couldn't allocate overlay surface: %s\n", SDL_GetError());
    exit(1);
  }

  ov_rect.x=0;
  ov_rect.y=0;
  ov_rect.w=conf.res_w;
  ov_rect.h=conf.res_h;
  init_rgb2yuv_table();
  // init_default_palette();
  init_message_yuv();
}

static __inline__ void draw_spr_yuy2(Uint32 *buf,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[0][sp->no_tile<<4];
  if (!sp->yflip) tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    if (c) {
      if (!(sp->priority)) buf[sp->x+nx]=rgb2yuv[grey[pal_obj[sp->pal][c]]].yuy2;
      else if (!back_col[sp->x+nx][CURLINE]) buf[sp->x+nx]=rgb2yuv[grey[pal_obj[sp->pal][c]]].yuy2;
    }
  }
}

static __inline__ void draw_obj_yuy2(Uint32 *buf)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) {
    draw_spr_yuy2(buf,&gb_spr[(Uint8)i]);
  }
}

static __inline__ void draw_back_yuy2(Uint32 *buf)
{
  Uint8 *tb,*tp;
  int y,x,i;
  int sx,sy;
  Uint8 /*bit0,bit1,*/c;
  Sint16 no_tile;
 
  if (LCDCCONT&0x08) tb=&vram_page[0][0x1c00];
  else  tb=&vram_page[0][0x1800];

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;

  no_tile=tb[i&0x1f];
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[0][no_tile<<4];
  tp+=(sy&0x07)<<1;
  
  x=0;
  for(sx=SCRX&0x07;sx<8;sx++,x++) {
    /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
      bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(sx);
    /* FIXME VIC VIPER */
    buf[x]=rgb2yuv[grey[gblcdc->vram_pal_line[x][c]]].yuy2;
    back_col[x][CURLINE]=c;
  }
  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    tp=&vram_page[0][no_tile<<4];
    tp+=(sy&0x07)<<1;
    for(sx=0;sx<8 && x+sx<160;sx++) {
      /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(sx);
      /* FIXME VIC VIPER */
      buf[x+sx]=rgb2yuv[grey[gblcdc->vram_pal_line[x+sx][c]]].yuy2;
      back_col[x+sx][CURLINE]=c;
    }
  }
}

static __inline__ void draw_win_yuy2(Uint32 *buf)
{
  Uint8 *tb,*tp;
  Sint16 y,x,i,sx=0;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c;
 
  if (LCDCCONT&0x40) tb=&vram_page[0][0x1c00];
  else tb=&vram_page[0][0x1800];
  
  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07)<<1;
    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[0][no_tile<<4];
      tp+=y;
      for(sx=0;sx<8 && (x+sx)<160;sx++) {
	/*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(sx);
	/* FIXME VIC VIPER */
	buf[x+sx]=rgb2yuv[grey[gblcdc->vram_pal_line[x+sx][c]]].yuy2;
	back_col[x+sx][CURLINE]=c;
      }
    }
  }  
}


static __inline__ void draw_spr_col_yuy2(Uint32 *buf,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[sp->page][sp->no_tile<<4];
  if (!sp->yflip) tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);

    /* TODO: OPTIMISATION */
    /*if (back_col[sp->x+nx][CURLINE]&0x80) {  
      if (c) {
	if (!(back_col[sp->x+nx][CURLINE]&0x0f)) 
	  buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      }
    } else {
      if (!sp->priority) {
	if (c) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      } else {
	if (!(back_col[sp->x+nx][CURLINE]&0x0f)) buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      }      
      }*/

    if (LCDCCONT&0x01) {
      if (c) {
	if (back_col[sp->x+nx][CURLINE]&0x80) {
	  if (!(back_col[sp->x+nx][CURLINE]&0x0f)) 
	    buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].yuy2;
 	} else if (!sp->priority) buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].yuy2;
	else if (!(back_col[sp->x+nx][CURLINE]&0x0f)) buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].yuy2;
      } 
    } else if (c) buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].yuy2;
  }
}


static __inline__ void draw_obj_col_yuy2(Uint32 *buf)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) 
    draw_spr_col_yuy2(buf,&gb_spr[(Uint8)i]);
}

static __inline__ void draw_back_col_yuy2(Uint32 *buf)
{
  Uint8 *tb,*tp,*att_tb;
  Uint16 y,x,i;
  Uint16 sx,sy;
  Uint8 /*bit0,bit1,*/c,p,att,xflip,yflip;
  Sint16 no_tile;
 
  if (LCDCCONT&0x08) {// select Tile Map
    tb=&vram_page[0][0x1c00];
    att_tb=&vram_page[1][0x1c00];
  } else {
    tb=&vram_page[0][0x1800];
    att_tb=&vram_page[1][0x1800];
  }

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  att_tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;
  
  no_tile=tb[i&0x1f];
  att=att_tb[i&0x1f];
  
  yflip = (att & 0x40) ? 1 : 0;
  xflip = (att & 0x20) ? 1 : 0;
  p=att&0x07;
   
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[(att&0x08)>>3][(no_tile<<4)];
  
  if (!yflip) tp+=((sy&0x07)<<1);
  else tp+=((7-(sy&0x07))<<1);

  x=0;
  for(sx=SCRX&0x07;sx<8;sx++,x++) {
    int wbit;
    if (!xflip) wbit=sx;
    else wbit=7-sx;
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    buf[x]=rgb2yuv[pal_col_bck[p][c]].yuy2;
    back_col[x][CURLINE]=(att&0x80)+c;
  }

  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    att=att_tb[i&0x1f];
    yflip = (att & 0x40) ? 1 : 0;
    xflip = (att & 0x20) ? 1 : 0;    
    p=att&0x07;

    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    
    tp=&vram_page[(att&0x08)>>3][(no_tile<<4)];
    if (!yflip) tp+=((sy&0x07)<<1);
    else tp+=((7-(sy&0x07))<<1);

    for(sx=0;sx<8;sx++) {
      int wbit;
      if (!xflip) wbit=sx;
      else wbit=7-sx;
      /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(wbit);
      buf[x+sx]=rgb2yuv[pal_col_bck[p][c]].yuy2;
      back_col[x+sx][CURLINE]=(att&0x80)+c;
    }
  }
}

static __inline__ void draw_win_col_yuy2(Uint32 *buf)
{
  Uint8 *tb,*tp,*att_tb; // ,*tiles;
  Uint16 y,x,i,sx;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c,p,att,xflip,yflip;
  
  if (LCDCCONT&0x40) {
    tb=&vram_page[0][0x1c00];
    att_tb=&vram_page[1][0x1c00];
  } else {
    tb=&vram_page[0][0x1800];
    att_tb=&vram_page[1][0x1800];
  }

  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    att_tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07);

    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      att=att_tb[i];
      yflip = (att & 0x40) ? 1 : 0;
      xflip = (att & 0x20) ? 1 : 0;  
      p=att&0x07;

      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[(att&0x08)>>3][(no_tile<<4)]; 
      if (!yflip) tp+=y<<1;
      else tp+=((7-y)<<1);
      
      for(sx=0;sx<8;sx++) {
	int wbit;
	if (!xflip) wbit=sx;
	else wbit=7-sx;
	/*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	  bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(wbit);
	buf[x+sx]=rgb2yuv[pal_col_bck[p][c]].yuy2;
	if (!(att&0x80)) back_col[x+sx][CURLINE]=c;
	else back_col[x+sx][CURLINE]=0x80+c;	    
      }
    }
  }  
}


void draw_screen_wb_yuy2(void)
{ 
  Uint32 *buf=(Uint32 *)overlay->pixels[0] + CURLINE*SCREEN_X;
#ifndef DEBUG
  if (LCDCCONT&0x01) draw_back_yuy2(buf);
  if (LCDCCONT&0x20) draw_win_yuy2(buf);
  if (LCDCCONT&0x02) draw_obj_yuy2(buf);
#else
  if (active_back && LCDCCONT&0x01) draw_back_yuy2(buf);
  if (active_win &&  LCDCCONT&0x20) draw_win_yuy2(buf);
  if (active_obj && LCDCCONT&0x02) draw_obj_yuy2(buf);
#endif
}

void draw_screen_col_yuy2(void) 
{
  Uint32 *buf=(Uint32 *)overlay->pixels[0] + CURLINE*SCREEN_X;
#ifndef DEBUG
  draw_back_col_yuy2(buf);
  if (LCDCCONT&0x20) draw_win_col_yuy2(buf);
  if (LCDCCONT&0x02) draw_obj_col_yuy2(buf);
#else
  if (active_back) draw_back_col_yuy2(buf);
  if (active_win &&  LCDCCONT&0x20) draw_win_col_yuy2(buf);
  if (active_obj && LCDCCONT&0x02) draw_obj_col_yuy2(buf);
#endif
}


VIDEO_MODE video_yuy2={
  init_video_yuy2,
  reinit_video_yuv,
  draw_screen_col_yuy2,
  draw_screen_wb_yuy2,
  NULL,
  blit_screen_yuv,
  NULL,
  draw_message_yuy2,
  NULL
};
