/*  gngb, a game boy color emulator
 *  Copyright (C) 2001 Peponas Thomas & Peponas Mathieu
 * 
 *  This program is free software; you can redistribute it and/or modify  
 *  it under the terms of the GNU General Public License as published by   
 *  the Free Software Foundation; either version 2 of the License, or    
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <SDL.h>
#include "global.h"
#include "memory.h"
#include "video_yuv.h"
#include "vram.h"
#include "message.h"
#include "tiny_font.c"
#include "emu.h"
#include "interrupt.h"

#ifdef DEBUG
#include "gngb_debuger/debuger.h"
#endif

__inline__ void draw_char_yv12(int ox,int oy,unsigned char c) {
  int fx,fy;
  int indice=c-32;
  int x,y;
  Uint8 *buf=(Uint8 *)overlay->pixels[0] + oy*overlay->pitches[0];
  Uint16 *bufy=(Uint16 *)overlay->pixels[0] + oy*overlay->pitches[0];
  Uint8 *bufu=(Uint8 *)overlay->pixels[1] + oy*overlay->pitches[1];
  Uint8 *bufv=(Uint8 *)overlay->pixels[2] + oy*overlay->pitches[2];
  
 
  fx=indice*wl;
  fy=0;
  if (conf.gb_type&COLOR_GAMEBOY) {
    for(y=0;y<hl*2;y+=2,bufy+=(overlay->pitches[0]-2)) {
      for(x=0;x<wl;x+=1) {
	static Uint8 Y;
	Y=tiny_font.pixel_data[((x+fx)+((y>>1)+fy)*tiny_font.width)*3];
	bufy[y+ox+x]=Y<<8|Y;
	bufy[y+ox+x+(overlay->pitches[0]>>1)]=Y<<8|Y;
      }
    }
    for(y=0;y<hl;y++,bufu+=overlay->pitches[1]-1,bufv+=overlay->pitches[2]-1)
      for(x=0;x<wl;x+=1) {
	bufu[y+ox+x]=128;
	bufv[y+ox+x]=128;
      }
  } else {
    for(y=0;y<hl;y++,buf+=overlay->pitches[0]-1)
      for(x=0;x<wl;x+=1) {
	buf[y+ox+x]=tiny_font.pixel_data[((x+fx)+(y+fy)*tiny_font.width)*3];
      }
  }
}


void draw_message_yv12(int x,int y,char *mes) {
  int i;
  /*  if (conf.gb_type&COLOR_GAMEBOY) y*=2;*/
  for(i=0;i<strlen(mes);i++)
    draw_char_yv12(x+i*wl,y,mes[i]);
}

#define COL32_TO_16(col) ((((col&0xff0000)>>19)<<11)|(((col&0xFF00)>>10)<<5)|((col&0xFF)>>3))

void set_pal_yv12(int i) 
{
  Uint16 u,v;
  grey[0]=rgb2yuv[COL32_TO_16(conf.pal[i][0])].y>>8;
  grey[1]=rgb2yuv[COL32_TO_16(conf.pal[i][1])].y>>8;
  grey[2]=rgb2yuv[COL32_TO_16(conf.pal[i][2])].y>>8;
  grey[3]=rgb2yuv[COL32_TO_16(conf.pal[i][3])].y>>8;
  
  u=rgb2yuv[COL32_TO_16(conf.pal[i][0])].u;
  v=rgb2yuv[COL32_TO_16(conf.pal[i][0])].v;
  u+=rgb2yuv[COL32_TO_16(conf.pal[i][1])].u;
  v+=rgb2yuv[COL32_TO_16(conf.pal[i][1])].v;
  u+=rgb2yuv[COL32_TO_16(conf.pal[i][2])].u;
  v+=rgb2yuv[COL32_TO_16(conf.pal[i][2])].v;
  u+=rgb2yuv[COL32_TO_16(conf.pal[i][3])].u;
  v+=rgb2yuv[COL32_TO_16(conf.pal[i][3])].v;
  u=u>>2;
  v=v>>2;
  
  memset(overlay->pixels[1],u,overlay->pitches[1]*(overlay->h/2));
  memset(overlay->pixels[2],v,overlay->pitches[2]*(overlay->h/2));
}

void init_video_yv12(Uint32 flag)
{
  yuv_flag=SDL_HWSURFACE|flag|SDL_RESIZABLE; /* YUV mode are resizable */

  gb_screen=SDL_SetVideoMode(conf.res_w,conf.res_h,BIT_PER_PIXEL,yuv_flag);
  
  if (conf.gb_type&COLOR_GAMEBOY) {
    overlay=SDL_CreateYUVOverlay(SCREEN_X*2,SCREEN_Y*2,SDL_YV12_OVERLAY,gb_screen); // FIXME : SGB support
  } else {
    overlay=SDL_CreateYUVOverlay(SCREEN_X,SCREEN_Y,SDL_YV12_OVERLAY,gb_screen); // FIXME : SGB support
  }
  /* Error check */
  if (gb_screen==NULL) {
    printf("Couldn't set %dx%dx%d video mode: %s\n",
	   SCREEN_X,SCREEN_Y,BIT_PER_PIXEL,SDL_GetError());
    exit(1);
  }
  if (overlay==NULL) {
    printf("Couldn't allocate overlay surface: %s\n", SDL_GetError());
    exit(1);
  }

  ov_rect.x=0;
  ov_rect.y=0;
  ov_rect.w=conf.res_w;
  ov_rect.h=conf.res_h;
  init_rgb2yuv_table();
  //  init_default_palette();
  init_message_yuv();
}


static __inline__ void draw_spr_yv12(Uint8 *buf,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[0][sp->no_tile<<4];
  if (!sp->yflip) tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    if (c) {
      if (!(sp->priority)) buf[sp->x+nx]=grey[pal_obj[sp->pal][c]];
      else if (!back_col[sp->x+nx][CURLINE]) buf[sp->x+nx]=grey[pal_obj[sp->pal][c]];
    }
  }
}

static __inline__ void draw_obj_yv12(Uint8 *buf)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) {
    draw_spr_yv12(buf,&gb_spr[(Uint8)i]);
  }
}

static __inline__ void draw_back_yv12(Uint8 *buf)
{
  Uint8 *tb,*tp;
  int y,x,i;
  int sx,sy;
  Uint8 /*bit0,bit1,*/c;
  Sint16 no_tile;
 
  if (LCDCCONT&0x08) tb=&vram_page[0][0x1c00];
  else  tb=&vram_page[0][0x1800];

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;

  no_tile=tb[i&0x1f];
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[0][no_tile<<4];
  tp+=(sy&0x07)<<1;
  
  x=0;
  for(sx=SCRX&0x07;sx<8;sx++,x++) {
    /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
      bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(sx);
    /* FIXME VIC VIPER */
    buf[x]=grey[gblcdc->vram_pal_line[x][c]];
    back_col[x][CURLINE]=c;
  }
  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    tp=&vram_page[0][no_tile<<4];
    tp+=(sy&0x07)<<1;
    for(sx=0;sx<8 && x+sx<160;sx++) {
      /*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(sx);
      /* FIXME VIC VIPER */
      buf[x+sx]=grey[gblcdc->vram_pal_line[x+sx][c]];
      back_col[x+sx][CURLINE]=c;
    }
  }
}

static __inline__ void draw_win_yv12(Uint8 *buf)
{
  Uint8 *tb,*tp;
  Sint16 y,x,i,sx=0;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c;
 
  if (LCDCCONT&0x40) tb=&vram_page[0][0x1c00];
  else tb=&vram_page[0][0x1800];
  
  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07)<<1;
    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[0][no_tile<<4];
      tp+=y;
      for(sx=0;sx<8 && (x+sx)<160;sx++) {
	/*bit0=((*tp)&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  bit1=((*(tp+1))&tab_ms[sx].mask)>>tab_ms[sx].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(sx);
	/* FIXME VIC VIPER */
	buf[x+sx]=grey[gblcdc->vram_pal_line[x+sx][c]];
	back_col[x+sx][CURLINE]=c;
      }
    }
  }  
}

static __inline__ void draw_spr_col_yv12(Uint16 *buf,Uint8 *buf_u, Uint8 *buf_v,GB_SPRITE *sp)
{
  Uint8 *tp;
  Uint8 nx;
  Uint8 /*bit0,bit1,*/wbit,c;

  tp=&vram_page[sp->page][sp->no_tile<<4];
  if (!sp->yflip) tp+=((sp->yoff)<<1);
  else tp+=(sp->sizey-1-sp->yoff)<<1;
  
  for(nx=sp->xoff;nx<8;nx++) {
    if (!sp->xflip) wbit=nx;
    else wbit=7-nx;	
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);

    /* TODO: OPTIMISATION */
    /*if (back_col[sp->x+nx][CURLINE]&0x80) {  // bg priority
      if (c) {
      if (!(back_col[sp->x+nx][CURLINE]&0x0f)) {
      buf[sp->x+nx]=rgb2y[pal_col_obj[sp->pal_col][c]];
      buf_u[sp->x+nx]=rgb2u[pal_col_obj[sp->pal_col][c]];
      buf_v[sp->x+nx]=rgb2v[pal_col_obj[sp->pal_col][c]];
      }
      }
      } else {
      if (!sp->priority) {
      if (c) {
      buf[sp->x+nx]=rgb2y[pal_col_obj[sp->pal_col][c]];
      buf_u[sp->x+nx]=rgb2u[pal_col_obj[sp->pal_col][c]];
      buf_v[sp->x+nx]=rgb2v[pal_col_obj[sp->pal_col][c]];
      }
      } else {
      if (!(back_col[sp->x+nx][CURLINE]&0x0f)) {
      buf[sp->x+nx]=rgb2y[pal_col_obj[sp->pal_col][c]];
      buf_u[sp->x+nx]=rgb2u[pal_col_obj[sp->pal_col][c]];
      buf_v[sp->x+nx]=rgb2v[pal_col_obj[sp->pal_col][c]];
      }
      }      
      }*/

    if (LCDCCONT&0x01) {
      if (c) {
	if (back_col[sp->x+nx][CURLINE]&0x80) {
	  if (!(back_col[sp->x+nx][CURLINE]&0x0f)) {
	    //buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
	    buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].y;
	    buf_u[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].u;
	    buf_v[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].v;
	  }
 	} else if (!sp->priority) {
	  //buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
	  buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].y;
	  buf_u[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].u;
	  buf_v[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].v;
	}
	else if (!(back_col[sp->x+nx][CURLINE]&0x0f)) {
	  //buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
	  buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].y;
	  buf_u[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].u;
	  buf_v[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].v;
	}
      }
    } else if (c) {
      //buf[sp->x+nx]=pal_col_obj[sp->pal_col][c];
      buf[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].y;
      buf_u[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].u;
      buf_v[sp->x+nx]=rgb2yuv[pal_col_obj[sp->pal_col][c]].v;
    }

  }
}


static __inline__ void draw_obj_col_yv12(Uint16 *buf,Uint8 *buf_u, Uint8 *buf_v)
{
  Sint8 i;
  for(i=nb_spr-1;i>=0;i--) 
    draw_spr_col_yv12(buf,buf_u,buf_v,&gb_spr[(Uint8)i]);
}

static __inline__ void draw_back_col_yv12(Uint16 *buf,Uint8 *buf_u, Uint8 *buf_v)
{
  Uint8 *tb,*tp,*att_tb;
  Uint16 y,x,i;
  Uint16 sx,sy;
  Uint8 /*bit0,bit1,*/c,p,att,xflip,yflip;
  Sint16 no_tile;
 
  if (LCDCCONT&0x08) {// select Tile Map
    tb=&vram_page[0][0x1c00];
    att_tb=&vram_page[1][0x1c00];
  } else {
    tb=&vram_page[0][0x1800];
    att_tb=&vram_page[1][0x1800];
  }

  y=CURLINE;
  sy=SCRY+CURLINE;
  tb+=((sy>>3)<<5)&0x3ff;
  att_tb+=((sy>>3)<<5)&0x3ff;
  i=SCRX>>3;
  
  no_tile=tb[i&0x1f];
  att=att_tb[i&0x1f];
  
  yflip = (att & 0x40) ? 1 : 0;
  xflip = (att & 0x20) ? 1 : 0;
  p=att&0x07;

   
  if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
  tp=&vram_page[(att&0x08)>>3][(no_tile<<4)];
  
  if (!yflip) tp+=((sy&0x07)<<1);
  else tp+=((7-(sy&0x07))<<1);

  x=0;
  for(sx=SCRX&0x07;sx<8 && sx<160;sx++,x++) {
    int wbit;
    if (!xflip) wbit=sx;
    else wbit=7-sx;
    /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
      c=(bit1<<1)|bit0;*/
    c=GET_GB_PIXEL(wbit);
    buf[x]=rgb2yuv[pal_col_bck[p][c]].y;
    buf_u[x]=rgb2yuv[pal_col_bck[p][c]].u;
    buf_v[x]=rgb2yuv[pal_col_bck[p][c]].v;
    back_col[x][CURLINE]=(att&0x80)+c;
  }

  i++;
  for(;x<160;x+=8,i++) {
    no_tile=tb[i&0x1f];
    att=att_tb[i&0x1f];
    yflip = (att & 0x40) ? 1 : 0;
    xflip = (att & 0x20) ? 1 : 0;    
    p=att&0x07;

    if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
    
    tp=&vram_page[(att&0x08)>>3][(no_tile<<4)];
    if (!yflip) tp+=((sy&0x07)<<1);
    else tp+=((7-(sy&0x07))<<1);

    for(sx=0;sx<8 && (x+sx)<160;sx++) {
      int wbit;
      if (!xflip) wbit=sx;
      else wbit=7-sx;
      /*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	c=(bit1<<1)|bit0;*/
      c=GET_GB_PIXEL(wbit);
      buf[x+sx]=rgb2yuv[pal_col_bck[p][c]].y;
      buf_u[x+sx]=rgb2yuv[pal_col_bck[p][c]].u;
      buf_v[x+sx]=rgb2yuv[pal_col_bck[p][c]].v;
      back_col[x+sx][CURLINE]=(att&0x80)+c;
    }
  }
}

static __inline__ void draw_win_col_yv12(Uint16 *buf,Uint8 *buf_u, Uint8 *buf_v)
{
  Uint8 *tb,*tp,*att_tb; // ,*tiles;
  Uint16 y,x,i,sx;
  Sint16 no_tile;
  Uint8 /*bit0,bit1,*/c,p,att,xflip,yflip;
  
  if (LCDCCONT&0x40) {
    tb=&vram_page[0][0x1c00];
    att_tb=&vram_page[1][0x1c00];
  } else {
    tb=&vram_page[0][0x1800];
    att_tb=&vram_page[1][0x1800];
  }

  if (WINX>=166) return;
  
  if (CURLINE>=WINY) {
    
    tb+=(((CURLINE-WINY)>>3)<<5);
    att_tb+=(((CURLINE-WINY)>>3)<<5);
    y=((CURLINE-WINY)&0x07);

    x=(((WINX-7)<0)?0:(WINX-7));
    for(i=0;x<160;x+=8,i++) {
      no_tile=tb[i];
      att=att_tb[i];
      yflip = (att & 0x40) ? 1 : 0;
      xflip = (att & 0x20) ? 1 : 0;  
      p=att&0x07;

      if (!(LCDCCONT&0x10)) no_tile=256+(signed char)no_tile;
      tp=&vram_page[(att&0x08)>>3][(no_tile<<4)]; 
      if (!yflip) tp+=y<<1;
      else tp+=((7-y)<<1);
      
      for(sx=0;sx<8;sx++) {
	int wbit;
	if (!xflip) wbit=sx;
	else wbit=7-sx;
	/*bit0=((*tp)&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	  bit1=((*(tp+1))&tab_ms[wbit].mask)>>tab_ms[wbit].shift;
	  c=(bit1<<1)|bit0;*/
	c=GET_GB_PIXEL(wbit);
	buf[x+sx]=rgb2yuv[pal_col_bck[p][c]].y;
	buf_u[x+sx]=rgb2yuv[pal_col_bck[p][c]].u;
	buf_v[x+sx]=rgb2yuv[pal_col_bck[p][c]].v;
	back_col[x+sx][CURLINE]=((att&0x80)+c);
      }
    }
  }  
}

void draw_screen_wb_yv12(void)
{
  Uint8 *buf=(Uint8 *)overlay->pixels[0] + CURLINE*SCREEN_X;
#ifndef DEBUG
  if (LCDCCONT&0x01) draw_back_yv12(buf);
  if (LCDCCONT&0x20) draw_win_yv12(buf);
  if (LCDCCONT&0x02) draw_obj_yv12(buf);
#else
  if (active_back && LCDCCONT&0x01) draw_back_yv12(buf);
  if (active_win &&  LCDCCONT&0x20) draw_win_yv12(buf);
  if (active_obj && LCDCCONT&0x02) draw_obj_yv12(buf);
#endif
}

static __inline__ void copy_yuv_line(Uint32 *src,Uint32 *dst) {
  int i;
  for (i=0;i<(overlay->pitches[0]>>2);i+=4) {
    dst[i]=src[i];
    dst[i+1]=src[i+1];
    dst[i+2]=src[i+2];
    dst[i+3]=src[i+3];
  }
}

void draw_screen_col_yv12(void) 
{
  Uint16 *buf=(Uint16 *)overlay->pixels[0] + CURLINE*overlay->pitches[0];
  Uint8  *buf_u=(Uint8 *)overlay->pixels[1] + CURLINE*overlay->pitches[1];
  Uint8  *buf_v=(Uint8 *)overlay->pixels[2] + CURLINE*overlay->pitches[2];
  //Uint16 yuv_interline_int=conf.yuv_interline_int;
#ifndef DEBUG
  draw_back_col_yv12(buf,buf_u,buf_v);
  if (LCDCCONT&0x20) draw_win_col_yv12(buf,buf_u,buf_v);
  if (LCDCCONT&0x02) draw_obj_col_yv12(buf,buf_u,buf_v);
#else
  if (active_back) draw_back_col_yv12(buf,buf_u,buf_v);
  if (active_win && LCDCCONT&0x20) draw_win_col_yv12(buf,buf_u,buf_v);
  if (active_obj && LCDCCONT&0x02) draw_obj_col_yv12(buf,buf_u,buf_v);
#endif
  
#if 0
  buf+=overlay->pitches[0]/2;
  /* Erase some trash made by drawback col (Better solution?) */
  buf[0]=yuv_interline_int;buf[1]=yuv_interline_int;buf[2]=yuv_interline_int;buf[3]=yuv_interline_int;
  buf[4]=yuv_interline_int;buf[5]=yuv_interline_int;buf[6]=yuv_interline_int;buf[7]=yuv_interline_int;
#else 
  copy_yuv_line((Uint32 *)buf,(Uint32 *)(buf+(overlay->pitches[0]>>1)));
#endif
}

VIDEO_MODE video_yv12={
  init_video_yv12,
  reinit_video_yuv,
  draw_screen_col_yv12,
  draw_screen_wb_yv12,
  NULL,
  blit_screen_yuv,
  set_pal_yv12,
  draw_message_yv12,
  NULL
};
